/**
 * 
 */
package com.datastax.hectorjpa.index;

import static com.datastax.hectorjpa.serializer.CompositeUtils.getCassType;
import static com.datastax.hectorjpa.serializer.CompositeUtils.newComposite;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Comparator;

import me.prettyprint.cassandra.serializers.ByteBufferSerializer;
import me.prettyprint.cassandra.serializers.BytesArraySerializer;
import me.prettyprint.cassandra.serializers.DynamicCompositeSerializer;
import me.prettyprint.cassandra.serializers.StringSerializer;
import me.prettyprint.hector.api.Keyspace;
import me.prettyprint.hector.api.beans.AbstractComposite.ComponentEquality;
import me.prettyprint.hector.api.beans.DynamicComposite;
import me.prettyprint.hector.api.mutation.Mutator;

import org.apache.cassandra.utils.ByteBufferUtil;
import org.apache.openjpa.kernel.OpenJPAStateManager;
import org.apache.openjpa.meta.FieldMetaData;
import org.apache.openjpa.util.MetaDataException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.datastax.hectorjpa.meta.key.KeyStrategy;
import com.datastax.hectorjpa.query.IndexQuery;
import com.datastax.hectorjpa.query.QueryIndexField;
import com.datastax.hectorjpa.query.QueryOrderField;
import com.datastax.hectorjpa.query.field.FieldExpression;
import com.datastax.hectorjpa.query.iterator.ScanBuffer;
import com.datastax.hectorjpa.service.IndexQueue;
import com.datastax.hectorjpa.store.CassandraClassMetaData;
import com.datastax.hectorjpa.store.MappingUtils;

/**
 * Class to perform all operations for secondary indexing on an instance in the
 * statemanager
 * 
 * @author Todd Nine
 * 
 */
public abstract class AbstractIndexOperation {
  

  private static final Logger logger = LoggerFactory.getLogger(AbstractIndexOperation.class);
 

  public static final String CF_NAME = "Index_Container";
  
  /**
   * The version to prepend to every row key for indexing
   */
  public static final int INDEXING_VERSION = 3;

  protected static byte[] HOLDER = new byte[] { 0 };

  protected static final StringSerializer stringSerializer = StringSerializer
      .get();

  protected static final DynamicCompositeSerializer compositeSerializer = new DynamicCompositeSerializer();

  protected static final BytesArraySerializer bytesSerializer = BytesArraySerializer
      .get();
  
  protected static final ByteBufferSerializer buffSerializer = ByteBufferSerializer.get();

  protected static int MAX_SIZE = 500;

  /**
   * The index name as a string.  Not really used during runtime other than for debugging output
   */
  protected String searchIndexNameString;
  
  protected String reverseIndexNameString;

  protected QueryIndexField[] fields;

  protected QueryOrderField[] orders;

  protected IndexDefinition indexDefinition;

  protected KeyStrategy keyStrategy;
  
  protected String compositeComparator;

  public AbstractIndexOperation(CassandraClassMetaData metaData,
      IndexDefinition indexDef) {
    this.indexDefinition = indexDef;

    IndexField[] fieldDirections = indexDef.getIndexedFields();
    IndexOrder[] indexOrders = indexDef.getOrderFields();

    this.fields = new QueryIndexField[fieldDirections.length];
    this.orders = new QueryOrderField[indexOrders.length];

    StringBuffer searchIndexName = new StringBuffer();
    StringBuffer reverseIndexName = new StringBuffer();
   
    FieldMetaData fmd = null;

    for (int i = 0; i < fieldDirections.length; i++) {
      fmd = metaData.getField(fieldDirections[i].getName());

      if (fmd == null) {
        throw new MetaDataException(
            String
                .format(
                    "You specified field '%s' as an index field, yet it does not exist in class '%s'",
                    fieldDirections[i].getName(), metaData.getDescribedType()));
      }

      fields[i] = new QueryIndexField(fmd);

      searchIndexName.append(fieldDirections[i].getName()).append("_");
      reverseIndexName.append(fieldDirections[i].getName()).append("_");
      

    }

    for (int i = 0; i < indexOrders.length; i++) {
      fmd = metaData.getField(indexOrders[i].getName());

      if (fmd == null) {
        throw new MetaDataException(
            String
                .format(
                    "You specified field '%s' as an order field, yet it does not exist in class '%s'",
                    indexOrders[i].getName(), metaData.getDescribedType()));
      }

      orders[i] = new QueryOrderField(indexOrders[i], fmd);

      String order = indexOrders[i].isAscending() ? "Asc" : "Desc";
      
      searchIndexName.append(indexOrders[i].getName()).append(order).append("_");
      reverseIndexName.append(indexOrders[i].getName()).append(order).append("_");
    }

    searchIndexName.append("search");
    reverseIndexName.append("reverse");
    
    
    searchIndexNameString = searchIndexName.toString();
    reverseIndexNameString = reverseIndexName.toString();
    
 

    // now get our is serializer
    keyStrategy = MappingUtils.getKeyStrategy(metaData);

    // if the ID doesn't implement comparable, we can't compare our results
    if (!Comparable.class.isAssignableFrom(metaData.getPrimaryKeyFields()[0]
        .getDeclaredType())) {
      throw new MetaDataException(
          String
              .format(
                  "Ids for indexes objects must implement Comparable.  Field '%s' on class '%s' does not implement comparable",
                  metaData.getPrimaryKeyFields()[0].getName(),
                  metaData.getDescribedType()));
    }
    
    
    compositeComparator = getCassType(buffSerializer);

  }

  /**
   * Write the index
   * 
   * @param stateManager
   * @param mutator
   * @param clock
   */
  public abstract void writeIndex(OpenJPAStateManager stateManager,
      Mutator<ByteBuffer> mutator, long clock, IndexQueue queue);


  /**
   * Scan the given index query and add the results to the provided set. The set
   * comparator of the dynamic columns are compared via a tree comparator
   * 
   * @param query
   */
  public ScanBuffer scanIndex(IndexQuery query, Keyspace keyspace) {

    DynamicComposite startScan = newComposite();
    DynamicComposite endScan = newComposite();

    int length = fields.length;

    int last = length - 1;

    FieldExpression exp = null;

    for (int i = 0; i < last; i++) {

      exp = query.getExpression(this.fields[i].getMetaData());

      this.fields[i].addToComposite(startScan, i, exp.getStart(),
          ComponentEquality.EQUAL);
      this.fields[i].addToComposite(endScan, i, exp.getEnd(),
          ComponentEquality.EQUAL);
    }

    exp = query.getExpression(this.fields[last].getMetaData());

    // We can only write non 0 separators at the last value in our composite
    // type

    this.fields[last].addToComposite(startScan, last, exp.getStart(),
        query.getStartEquality());
    this.fields[last].addToComposite(endScan, last, exp.getEnd(),
        query.getEndEquality());

    return new ScanBuffer(keyspace, startScan, endScan, getScanKey());

    // super.executeQuery(startScan, endScan, results, keyspace);

  }
  
  /**
   * Return the row key to perform scanning on
   * @return
   */
  protected abstract ByteBuffer getScanKey();
 
  /**
   * Remove all values from the index that were for the given statemanager
   * 
   * @param stateManager
   * @param queue
   */
  public void removeIndexes(OpenJPAStateManager stateManager, IndexQueue queue,
      long clock) {

    ByteBuffer key = keyStrategy.toByteBuffer(stateManager.fetchObjectId());

    DynamicComposite composite = newComposite();

    composite.addComponent(key, buffSerializer);
    
    // queue the index values to be deleted
    queueDeletes(composite, queue, clock);

  }
  
  /**
   * Queue the read and delete operations for every index row in this object
   * @param searchCol
   * @param queue
   */
  protected abstract void queueDeletes(DynamicComposite searchCol, IndexQueue queue, long clock);

  /**
   * Construct the 2 composites from the fields in this index.
   * @param searchComposite
   * @param oldComposite
   * @return
   */
  protected void constructComposites(DynamicComposite searchComposite, DynamicComposite tombstoneComposite,
      DynamicComposite auditComposite, OpenJPAStateManager stateManager) {

    ByteBuffer key = keyStrategy.toByteBuffer(stateManager.fetchObjectId());

    Object field;

    tombstoneComposite.setComponent(0, key, buffSerializer, compositeComparator, ComponentEquality.EQUAL);

    auditComposite.setComponent(0, key, buffSerializer, 	compositeComparator, ComponentEquality.EQUAL);

    // now construct the composite with order by the ids at the end.
    for (QueryIndexField indexField : fields) {
      field = indexField.getValue(stateManager,
          stateManager.getPersistenceCapable());

      // add this to all deletes for the order composite.
      indexField.addFieldWrite(searchComposite, field);
      indexField.addFieldWrite(tombstoneComposite, field);

      
    }

    // now construct the composite with order by the ids at the end.
    for (AbstractIndexField order : orders) {

      // get the field value from our current index
      field = order
          .getValue(stateManager, stateManager.getPersistenceCapable());

      // add this to all deletes for the order composite.
      order.addFieldWrite(searchComposite, field);
      order.addFieldWrite(tombstoneComposite, field);

    }

    // add it to our new value

    searchComposite.addComponent(key, buffSerializer, compositeComparator);

    if(logger.isDebugEnabled()){
      logger.debug("Writing search index with row key: {} and column: {}", this.searchIndexNameString, ByteBufferUtil.bytesToHex(searchComposite.serialize()));
      logger.debug("Writing tombstone index with row key: {} and column: {}", this.reverseIndexNameString, ByteBufferUtil.bytesToHex(tombstoneComposite.serialize()));
    }
    

  }


  public Comparator<DynamicComposite> getComprator() {
    return new ResultComparator();
  }

  /**
   * @return the indexDefinition
   */
  public IndexDefinition getIndexDefinition() {
    return indexDefinition;
  }

  public class ResultComparator implements Comparator<DynamicComposite> {

    @SuppressWarnings("unchecked")
    @Override
    public int compare(DynamicComposite c1, DynamicComposite c2) {

      if (c1 == null && c2 != null) {
        return -1;
      }

      if (c2 == null && c1 != null) {
        return 1;
      }

      if (c1 == null && c2 == null) {
        return 0;
      }

      int compare = 0;

      int size = 0;

      Comparable<Object> c1Id = null;
      Comparable<Object> c2Id = null;

      // no order by, just order by each field starting from the beginning
      if (orders.length == 0) {

        size = fields.length;

        for (int i = 0; i < size; i++) {
          
          compare = fields[i].compare(c1, c2, i);
          
//          c1Id = (Comparable<Object>) c1.get(i, fields[i].getSerializer());
//
//          c2Id = (Comparable<Object>) c2.get(i, fields[i].getSerializer());
//
//          compare = c1Id.compareTo(c2Id);

          if (compare != 0) {
            return compare;
          }
        }
        
        ByteBuffer id1 = c1.get(c1.size() -1, buffSerializer);

        c1Id = (Comparable<Object>) keyStrategy.getInstance(id1);
        
        ByteBuffer id2 = c2.get(c2.size() -1, buffSerializer);

        c2Id = (Comparable<Object>) keyStrategy.getInstance(id2);

        return c1Id.compareTo(c2Id);

      }

      size = c1.getComponents().size();

      int c1StartIndex = size - orders.length - 1; // c1.getComponents().size()
     
//      int c2StartIndex = size - orders.length - 1; // c1.getComponents().size()


      for (int i = 0; i < orders.length; i++) {

        compare = orders[i].compare(c1, c2, c1StartIndex + i);

        if (compare != 0) {
          return compare;
        }
      }

      // if we get here the compare fields are equal. Compare ids
      ByteBuffer id1 = c1.get(c1.size() -1, buffSerializer);

      c1Id = (Comparable<Object>) keyStrategy.getInstance(id1);
      
      ByteBuffer id2 = c2.get(c2.size() -1, buffSerializer);

      c2Id = (Comparable<Object>) keyStrategy.getInstance(id2);
      

      return c1Id.compareTo(c2Id);

    }

  }
  
  /**
   * Create an index key in the values of version+className+indexName
   * @param className
   * @param indexName
   * @return
   */
  protected ByteBuffer createIndexKey(String className, String indexName){
    
    String indexVersion = String.format("%d_", INDEXING_VERSION);
    
    ByteBuffer indexKey = ByteBuffer.allocate(indexVersion.length()+className.length()+indexName.length());
    indexKey.put(StringSerializer.get().toByteBuffer(indexVersion));
    indexKey.put(StringSerializer.get().toByteBuffer(className));
    indexKey.put(StringSerializer.get().toByteBuffer(indexName));
    indexKey.rewind();
    return indexKey;
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append("fields=")
    .append(Arrays.asList(fields))
    .append(",indexDefinition=")
    .append(indexDefinition)
    .append(",keyStrategy=")
    .append(keyStrategy)
    .append(",orders=")
    .append(Arrays.asList(orders));
    return sb.toString();
  }
  
  

}
