package com.datastax.hectorjpa.meta.collection;

import static com.datastax.hectorjpa.serializer.CompositeUtils.getCassType;

import java.nio.ByteBuffer;
import java.util.Collection;

import me.prettyprint.cassandra.serializers.ByteBufferSerializer;
import me.prettyprint.cassandra.serializers.DynamicCompositeSerializer;
import me.prettyprint.cassandra.serializers.StringSerializer;
import me.prettyprint.hector.api.Keyspace;
import me.prettyprint.hector.api.beans.ColumnSlice;
import me.prettyprint.hector.api.beans.DynamicComposite;
import me.prettyprint.hector.api.factory.HFactory;
import me.prettyprint.hector.api.mutation.Mutator;
import me.prettyprint.hector.api.query.QueryResult;
import me.prettyprint.hector.api.query.SliceQuery;

import org.apache.openjpa.kernel.OpenJPAStateManager;
import org.apache.openjpa.meta.FieldMetaData;
import org.apache.openjpa.util.ChangeTracker;
import org.apache.openjpa.util.MetaDataException;
import org.apache.openjpa.util.Proxy;

import com.datastax.hectorjpa.meta.Field;
import com.datastax.hectorjpa.meta.key.KeyStrategy;
import com.datastax.hectorjpa.store.CassandraClassMetaData;
import com.datastax.hectorjpa.store.MappingUtils;

/**
 * Base class for all collection operations
 * 
 * @author Todd Nine
 * 
 */
public abstract class AbstractCollectionField extends Field {

  public static final String CF_NAME = "Collection_Container";
  
  /**
   * The version to prepend to every row key for indexing
   */
  public static final int COLLECTION_VERSION = 2;

  protected static ByteBuffer HOLDER = ByteBuffer.wrap(new byte[] { 0 });

  // the default batch size when it hasn't been set into the context
  protected int DEFAULT_FETCH_SIZE = Integer.MAX_VALUE;

  protected static final DynamicCompositeSerializer compositeSerializer = new DynamicCompositeSerializer();
  
  protected static final ByteBufferSerializer buffSerializer = ByteBufferSerializer.get();

//  protected final KeyStrategy ownerKeyStrategy;
  
  protected final KeyStrategy elementKeyStrategy;
  
  protected final Class<?> targetClass;

  // The name of this entity serialzied as bytes
  protected final ByteBuffer entityName;

  // the name of the field serialzied as bytes
  protected final ByteBuffer fieldName;
  

  protected int compositeFieldLength = 0;
  
  protected String compositeComparator;
  

  public AbstractCollectionField(FieldMetaData fmd) {
    super(fmd.getIndex(), fmd.getName());

    Class<?> clazz = fmd.getDeclaredType();

    if (!Collection.class.isAssignableFrom(clazz)) {
      throw new MetaDataException("Only collections are currently supported");
    }

    this.name = fmd.getName();

    CassandraClassMetaData elementClassMeta = (CassandraClassMetaData) fmd.getElement().getDeclaredTypeMetaData();
    
    if(elementClassMeta == null){
      throw new MetaDataException(String.format("You defined type %s in a collection, but it is not a persistable entity", fmd.getElement().getDeclaredType()));
    }
    
    elementKeyStrategy = MappingUtils.getKeyStrategy(elementClassMeta);

    // set the class of the collection elements
    targetClass = elementClassMeta.getDescribedType();

    // create our cached bytes for better performance
    fieldName = StringSerializer.get().toByteBuffer(name);

    // write our column family name of the owning side to our rowkey for
    // scanning
    String columnFamilyName = MappingUtils.getColumnFamily(elementClassMeta);

    entityName = StringSerializer.get().toByteBuffer(columnFamilyName);
    
    compositeComparator = getCassType(buffSerializer);
  }
  
  /**
   * Create our key byte array
   * 
   * @param entityIdBytes
   * @return
   */
  protected ByteBuffer constructKey(ByteBuffer entityIdBytes, ByteBuffer marker) {
    ByteBuffer buff = ByteBuffer.allocate(4 + entityName.remaining() + fieldName.remaining()
        + entityIdBytes.remaining() + marker.remaining());

    buff.putInt(COLLECTION_VERSION);
    buff.put(entityName.duplicate());
    buff.put(entityIdBytes.duplicate());
    buff.put(fieldName.duplicate());
    buff.put(marker.duplicate());
    buff.rewind();
    
    return buff;

  }

 
  /**
   * Create the slice query for this field
   * @param objectId
   * @param keyspace
   * @param columnFamilyName
   * @param count
   * @return
   */
  public SliceQuery<ByteBuffer, DynamicComposite, ByteBuffer> createQuery(ByteBuffer ownerKey,  Keyspace keyspace, int count) {
    
    //undefined value set it to something realistic
    if(count < 0){
      count = DEFAULT_FETCH_SIZE;
    }
    
    ByteBuffer rowKey = constructKey(ownerKey, getDefaultSearchmarker());
    
    SliceQuery<ByteBuffer, DynamicComposite, ByteBuffer> query = HFactory.createSliceQuery(keyspace, ByteBufferSerializer.get(), compositeSerializer,
        ByteBufferSerializer.get());

    query.setRange(null, null, false, count);
    query.setKey(rowKey);
    query.setColumnFamily(CF_NAME);
    return query;

  }
  

  /**
   * Useful for proxy fields.  The index 0 is the original value the index 1 is the new value.
   * If no proxy is present only index 1 will contain a value
   * 
   * @param value
   * @return
   */
  protected Object[] getDelta(Object value){
    Object[] values = new Object[2];
    
    if(!(value instanceof Proxy)){
      values[1] = value;
      return values;
    }
    
    ChangeTracker tracker = ((Proxy) value).getChangeTracker();
    
    Collection<?> objects = tracker.getRemoved();
    
    for(Object removed: objects){
      values[0] = removed;
      break;
    }

    objects = tracker.getAdded();
    
    for(Object added: objects){
      values[1] = added;
      break;
    }
    
    
    return values;
  
  }
  /**
   * Read this field from the data store from the queryresult
   * @param stateManager
   * @param result
   * @return true if this field had values
   */
  public abstract boolean readField(OpenJPAStateManager stateManager,   QueryResult<ColumnSlice<DynamicComposite, ByteBuffer>> result);
  
  
  /**
   * Remove this collection by removing the row key represending it
   * @param stateManager 
   * @param mutator
   * @param long clock
   * @param key The key of the stateManager's entity
   */
  public abstract void removeCollection(OpenJPAStateManager stateManager, Mutator<ByteBuffer> mutator, long clock, ByteBuffer key);
  
  /**
   * Return the default end bytes on the rowkey for searching an index
   * @return
   */
  protected abstract ByteBuffer getDefaultSearchmarker();
  
  

  @Override
  public String toString() {  
    return String.format("AbstractCollectionField(fieldId: %d, name: %s)", fieldId, name);
  }
 
  
  
  
  
}
